@@ -22,13 +22,14 @@
 #include <string.h>
 #include <strings.h>
 #include <errno.h>
+#include <ctype.h>
 
 #include "xdo.h"
 #include "xdotool.h"
 
-char *PROGRAM;
 static int script_main(int argc, char **argv);
 static int args_main(int argc, char **argv);
+int context_execute(context_t *context);
 void consume_args(context_t *context, int argc);
 void window_save(context_t *context, Window window);
 void window_list(context_t *context, const char *window_arg,
@@ -103,12 +104,15 @@ int window_get_arg(context_t *context, int min_arg, int window_arg_pos,
             context->argc, min_arg);
     return False;
   } else if (context->argc == min_arg) {
+    //fprintf(stderr, "Using default arg\n");
     /* nothing, keep default */
   } else if (context->argc > min_arg) {
     if (is_command(context->argv[min_arg])) {
+      //fprintf(stderr, "arg is command, using default\n");
       /* keep default */
     } else {
       /* got enough args, let's use the window you asked for */
+      //fprintf(stderr, "got enough args\n");
       *window_arg = context->argv[window_arg_pos];
       consume_args(context, 1);
     }
@@ -128,19 +132,19 @@ void window_list(context_t *context, const char *window_arg,
   /* If window_arg is NULL and we have windows in the list, use the list.
    * If window_arg is "%@" and we have windows in the list, use the list.
    * If window_arg is "%N" and we have windows in the list, use Nth window.
-   * Otheriwse, assume it's a window id.
+   *   'N' above must be a positive number.
+   * Otherwise, assume it's a window id.
+   *
+   * TODO(sissel): Not implemented yet:
+   * If window_arg is "%r" it means the root window of the current screen.
+   * If window_arg is "%q" it means we will wait for you to select a window
+   *   by clicking on it. (May not be necessary since we have 'selectwindow')
+   * If window_arg is "%c" it means the currently-active window.
    */
 
   *nwindows_ret = 0;
   *windowlist_ret = NULL;
 
-  //if (window_arg == NULL && context->nwindows > 0) {
-    /* Default is to use the first window matched */
-    //context->window_placeholder[0] = context->windows[0];
-    //*windowlist_ret = context->window_placeholder;
-    //*nwindows_ret = 1;
-  //if (window_arg == NULL) {
-  //} else
   if (window_arg != NULL && window_arg[0] == '%') {
     if (context->nwindows == 0) {
       fprintf(stderr, "There are no windows on the stack, Can't continue.\n");
@@ -159,7 +163,14 @@ void window_list(context_t *context, const char *window_arg,
     if (window_arg[1] == '@') {
       *windowlist_ret = context->windows;
       *nwindows_ret = context->nwindows;
+    } else if (window_arg[1] == 'q') {
+      /* TODO(sissel): Wait for you to click on the window. */
+    } else if (window_arg[1] == 'r') {
+      /* TODO(sissel): Get the root window of the current screen */
+    } else if (window_arg[1] == 'c') {
+      /* TODO(sissel): Get the current window */
     } else {
+      /* Otherwise assume %N */
       int window_index = atoi(window_arg + 1);
       if (window_index < 0) {
         /* negative offset */
@@ -208,14 +219,18 @@ struct dispatch {
   /* Query functions */
   { "getactivewindow", cmd_getactivewindow, },
   { "getwindowfocus", cmd_getwindowfocus, },
+  { "getwindowname", cmd_getwindowname, },
   { "getwindowpid", cmd_getwindowpid, },
   { "search", cmd_search, },
+  { "selectwindow", cmd_window_select, },
 
   /* Help me! */
   { "help", cmd_help, },
   { "version", cmd_version, },
 
   /* Action functions */
+  { "behave", cmd_behave, },
+  { "behave_screen_edge", cmd_behave_screen_edge, },
   { "click", cmd_click, },
   { "getmouselocation", cmd_getmouselocation, },
   { "key", cmd_key, },
@@ -225,23 +240,30 @@ struct dispatch {
   { "mousemove", cmd_mousemove, },
   { "mousemove_relative", cmd_mousemove_relative, },
   { "mouseup", cmd_mouseup, },
+  { "set_window", cmd_set_window, },
   { "type", cmd_type, },
   { "windowactivate", cmd_windowactivate, },
   { "windowfocus", cmd_windowfocus, },
+  { "windowkill", cmd_windowkill, },
   { "windowmap", cmd_windowmap, },
+  { "windowminimize", cmd_windowminimize, },
   { "windowmove", cmd_windowmove, },
   { "windowraise", cmd_windowraise, },
+  { "windowreparent", cmd_windowreparent, },
   { "windowsize", cmd_windowsize, },
   { "windowunmap", cmd_windowunmap, },
 
-  { "set_window", cmd_set_window, },
-
   { "set_num_desktops", cmd_set_num_desktops, },
   { "get_num_desktops", cmd_get_num_desktops, },
   { "set_desktop", cmd_set_desktop, },
   { "get_desktop", cmd_get_desktop, },
   { "set_desktop_for_window", cmd_set_desktop_for_window, },
   { "get_desktop_for_window", cmd_get_desktop_for_window, },
+  { "get_desktop_viewport", cmd_get_desktop_viewport, },
+  { "set_desktop_viewport", cmd_set_desktop_viewport, },
+
+  { "exec", cmd_exec, },
+
   { NULL, NULL, },
 };
 
@@ -249,39 +271,54 @@ int is_command(char* cmd) {
   int i;
   for (i = 0; dispatch[i].name != NULL; i++) {
       if (!strcasecmp(dispatch[i].name, cmd)) {
-	return 1;
+        return 1;
       }
     }
   return 0;
 }
 
 int main(int argc, char **argv) {
-  /* read stdin if stdin is not a tty or first argument is "-" */
+  return xdotool_main(argc, argv);
+}
+
+int xdotool_main(int argc, char **argv) {
+
+  /* If argv[1] is a file or "-", read commands from file or stdin,
+   * else use commands from argv.
+   */
 
-  int want_script;
   struct stat data;
   int stat_ret;
 
-  /* If we are being run from a script with shebang line #!/path/to/xdotool
-   * then the input file will be argv[argc - 1].
-   */
-  stat_ret = stat(argv[argc - 1], &data);
-
-  want_script = ((argc == 2 && !strcmp(argv[1], "-"))
-                 || (argc == 2 && stat_ret == 0));
-  if (want_script) {
-    return script_main(argc, argv);
-  } else {
-    return args_main(argc, argv);
+  if (argc >= 2) {
+    /* See if the first argument is an existing file */
+    stat_ret = stat(argv[1], &data);
+    
+    if ( strcmp(argv[1], "-") == 0 || stat_ret == 0) {
+      return script_main(argc, argv);
+    }
   }
+  return args_main(argc, argv);
 }
 
 int script_main(int argc, char **argv) {
+  /* Tokenize the input file while expanding positional parameters and
+   * environment variables. Pass the resulting argument list to
+   * args_main().
+   */
+
   FILE *input = NULL;
-  const char *path = argv[argc - 1];
-  char *cmd;
+  const char *path = argv[1];
   char buffer[4096];
 
+  char **script_argv = (char **) calloc(1, sizeof(char *));
+  int script_argc = 1;
+
+  /* copy argv[0] to script_argv[0] */
+  script_argv[0] = (char *) calloc(strlen(argv[0])+1, sizeof(char));
+  strcpy(script_argv[0], argv[0]);
+
+  /* determine whether reading from a file or from stdin */
   if (!strcmp(path, "-") || !isatty(0)) {
     input = fdopen(0, "r");
   } else {
@@ -291,42 +328,121 @@ int script_main(int argc, char **argv) {
       return EXIT_FAILURE;
     }
   }
+  
+  /* read input... */
+  int pos;
+  char *token;
 
-  int ret;
   while (fgets(buffer, 4096, input) != NULL) {
     char *line = buffer;
-    // Ignore leading whitespace
+    token = NULL;
+
+    /* Ignore leading whitespace */
     line += strspn(line, " \t");
 
-    // blanklines or line comment are ignored, too
+    /* blanklines or line comment are ignored, too */
     if (line[0] == '\n' || line[0] == '#') {
       continue;
     }
-    line[strlen(line) - 1] = '\0'; /* replace newline with null */
 
-    if (asprintf(&cmd, "%s %s", argv[0], line) == -1) {
-      fprintf(stderr, "asprintf failed\n");
-      exit(1);
-    }
-    //printf("Running: %s\n", cmd);
-    ret = system(cmd);
-    free(cmd);
+    /* replace newline with null */
+    if (line[strlen(line)-1] == '\n') 
+      line[strlen(line)-1] = '\0'; 
 
-    if (ret != EXIT_SUCCESS) {
-      return ret;
+    /* tokenize line into script_argv... */
+    while (strlen(line)) {
+      token = NULL;
+
+      /* modify line to contain the current token. Tokens are
+       * separated by whitespace, or quoted with single/double quotes.
+       */
+      if (line[0] == '"') {
+        line++;
+        line[strcspn(line, "\"")] = '\0';
+      }
+      else if (line[0] == '\'') {
+        line++;
+        line[strcspn(line, "\'")] = '\0';
+      }
+      else {
+        line[strcspn(line, " \t")] = '\0';
+      }
+
+      /* if a token begins with "$", append the corresponding
+       * positional parameter or environment variable to
+       * script_argv...
+      */
+      if (line[0] == '$') {
+        /* ignore dollar sign */
+        line++; 
+      
+        if (isdigit(line[0])) {
+          /* get the position of this parameter in argv */ 
+          pos = atoi(line) + 1; /* $1 is actually index 2 in the argv array */
+
+          /* bail if no argument was given for this parameter */
+          if (pos >= argc) {
+            fprintf (stderr, "%s: error: `%s' needs at least %d %s; only %d given\n",
+                     argv[0], argv[1], pos - 1, pos == 2 ? "argument" : "arguments",
+                     argc - 2);
+            return EXIT_FAILURE;
+          }
+          /* use command line argument */
+          token = argv[pos];
+        }
+        else {
+          /* use environment variable */
+          token = getenv(line);
+          if (token == NULL) {
+            /* since it's not clear what we should do if this env var is not
+             * present, let's abort */
+            fprintf(stderr, "%s: error: environment variable $%s is not set.\n",
+                    argv[0], line);
+            return EXIT_FAILURE;
+          }
+        }
+      }
+      else { 
+        /* use the verbatim token */
+        token = line;
+      }      
+
+      /* append token */
+      if (token != NULL) {
+
+        script_argv = realloc(script_argv, (script_argc+1) * sizeof(char *));
+        if (script_argv == NULL) {
+          fprintf(stderr, "%s: error: failed to allocate memory while parsing `%s'.\n", 
+                  argv[0], argv[1]);
+          exit(EXIT_FAILURE);
+        }
+        script_argv[script_argc] = (char *) calloc(strlen(token)+1, sizeof(char));
+
+        //printf("arg %d: %s\n", script_argc, token);
+        strncpy(script_argv[script_argc], token, strlen(token)+1);      
+        script_argc++;
+      }
+      
+      /* advance line to the next token */
+      line += strlen(line)+1;
+      line += strspn(line, " \t");
     }
   }
+  fclose(input);
 
-  return EXIT_SUCCESS;
+  /* Add NULL at the end */
+  script_argv = realloc(script_argv, (script_argc+1) * sizeof(char *));
+  /* TODO(sissel): STOPPED HERE */
+
+  /* run the parsed script */
+  return args_main(script_argc, script_argv);
 }
 
 int args_main(int argc, char **argv) {
-  char *cmd;
   int ret = 0;
-  int cmd_found = 0;
-  int i;
   int opt;
   int option_index;
+
   const char *usage = "Usage: %s <cmd> <args>\n";
   static struct option long_options[] = {
     { "help", no_argument, NULL, 'h' },
@@ -339,6 +455,10 @@ int args_main(int argc, char **argv) {
     cmd_help(NULL);
     exit(1);
   }
+  
+  //for(i = 0; i<argc; i++) {
+    //fprintf(stderr, "argv[%d] = \"%s\"\n", i, argv[i]);
+  //}
 
   while ((opt = getopt_long_only(argc, argv, "++hv", long_options, &option_index)) != -1) {
     switch (opt) {
@@ -353,47 +473,57 @@ int args_main(int argc, char **argv) {
         exit(EXIT_FAILURE);
     }
   }
-
-  PROGRAM = *argv;
-  argv++; argc--;
-
+  
   context_t context;
   context.xdo = xdo_new(NULL);
+  context.prog = *argv;
+  argv++; argc--;
   context.argc = argc;
   context.argv = argv;
   context.windows = NULL;
   context.nwindows = 0;
+  context.have_last_mouse = False;
+  context.debug = (getenv("DEBUG") != NULL);
 
   if (context.xdo == NULL) {
     fprintf(stderr, "Failed creating new xdo instance\n");
     return 1;
   }
 
+  ret = context_execute(&context);
+
+  xdo_free(context.xdo);
+  if (context.windows != NULL) {
+    free(context.windows);
+  }
+
+  return ret;
+} /* int args_main(int, char **) */
+
+int context_execute(context_t *context) {
+  int cmd_found = 0;
+  int i = 0;
+  char *cmd = NULL;
+  int ret = XDO_SUCCESS;
+
   /* Loop until all argv is consumed. */
-  while (context.argc > 0 && ret == XDO_SUCCESS) {
-    cmd = context.argv[0];
+  while (context->argc > 0 && ret == XDO_SUCCESS) {
+    cmd = context->argv[0];
     cmd_found = 0;
     for (i = 0; dispatch[i].name != NULL && !cmd_found; i++) {
       if (!strcasecmp(dispatch[i].name, cmd)) {
         cmd_found = 1;
         optind = 0;
-        ret = dispatch[i].func(&context);
+        ret = dispatch[i].func(context);
       }
     }
 
     if (!cmd_found) {
-      fprintf(stderr, "%s: Unknown command: %s\n", PROGRAM, cmd);
-      fprintf(stderr, "Run '%s help' if you want a command list\n", PROGRAM);
+      fprintf(stderr, "%s: Unknown command: %s\n", context->prog, cmd);
+      fprintf(stderr, "Run '%s help' if you want a command list\n", context->prog);
       ret = 1;
     }
   } /* while ... */
-
-  xdo_free(context.xdo);
-
-  if (context.windows != NULL) {
-    free(context.windows);
-  }
-
   return ret;
 } /* int args_main(int, char **) */
 
